package org.garret.scheduler;

import java.sql.*;
import java.io.*;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.ArrayList;
import java.util.Calendar;


public class Scheduler implements Runnable {
    public void run() {
        loadJobTable();
        restoreState();
        startQueuePollThread();
        startPermanentJobs();
        startPeriodicJobs();
        close();
    }
    
    public void terminate() {
        synchronized (stopEvent) { 
            queuePollInterval = 0;
            running = false;
            stopEvent.notifyAll();
        }
    }

    void close() {        
        for (Job job = jobList; job != null; job = job.next) { 
            job.close();
        }
        try { 
            connection.close();
        } catch (SQLException x) { 
            traceException("DATABASE-CLOSE", x);
        }
    }

    void startJob(String jobName) { 
        Job job = getJob(jobName);
        if (job != null) { 
            startJob(job);
        }
    }

    synchronized void startJob(Job job) { 
        if (job.state != Job.STATE_RUNNING) {
            job.state = Job.STATE_RUNNING;
            job.started = new Timestamp(System.currentTimeMillis());
            //job.terminated = null;
            job.terminated = new Timestamp(-1);
            job.restartCount = 0;
            String msg = "Job '" + job.name + " " + job.params + "' started at " + job.started;
            job.log.println(msg);
            if (verbose) { 
                message(msg);
            }
            try { 
                updateJobStmt.setInt(1, Job.STATE_RUNNING);
                updateJobStmt.setTimestamp(2, job.started);
                updateJobStmt.setTimestamp(3, job.terminated);
                updateJobStmt.setString(4, job.name);
                updateJobStmt.executeUpdate();
            } catch (SQLException x) { 
                traceException("START-JOB", x);
            }
            pool.start(job);
        }
    }

    synchronized void endJob(Job job, int state) { 
        job.state = state;
        job.terminated = new Timestamp(System.currentTimeMillis());
        try { 
            updateJobStmt.setInt(1, state);
            updateJobStmt.setTimestamp(2, job.started);
            updateJobStmt.setTimestamp(3, job.terminated);
            updateJobStmt.setString(4, job.name);
            updateJobStmt.executeUpdate();
        } catch (SQLException x) { 
            traceException("START-JOB", x);
        }
        String msg = "Job '" + job.name + " " + job.params
            + ((state == Job.STATE_FINISHED) ? "' finished at " :  "' failed at ")
            + job.terminated;        
        job.log.println(msg);
        if (verbose) { 
            message(msg);
        }
        if (state == Job.STATE_FINISHED) { 
            if (job.onFinish != null) { 
                for (int i = 0; i < job.onFinish.length; i++) {
                    startJob(job.onFinish[i]);
                }
            }
        } else { 
            if (job.onFailure != null) { 
                for (int i = 0; i < job.onFailure.length; i++) {
                    startJob(job.onFailure[i]);
                }
            }
        }
    }

    void startPermanentJobs() {
        for (Job job = jobList; job != null; job = job.next) { 
            if (job.schedulePolicy == Job.SCHED_PERMANENT) { 
                startJob(job);
            }
        }
    }

    void restoreState() {
        try { 
            Job job;
            Statement stmt = connection.createStatement();
            ResultSet cursor = stmt.executeQuery("select * from JobState");
            Job toBeRestartedList = null;
            while (cursor.next()) { 
                if (cursor.getInt("status") == Job.STATE_RUNNING) {
                    job = getJob(cursor.getString("jobname"));
                    if (job != null) { 
                        if (verbose) { 
                            message("Job '" + job.name + "' will be restarted");
                        }
                        job.nextToBeRestarted = toBeRestartedList;
                        toBeRestartedList = job;
                    }
                }
            }
            cursor.close();
            stmt.executeUpdate("delete from JobState");
            stmt.close();
            PreparedStatement pstmt = connection.prepareStatement("insert into JobState (jobname, status) values (?,0)");
            for (job = jobList; job != null; job = job.next) { 
                pstmt.setString(1, job.name);
                pstmt.executeUpdate();
            }
            pstmt.close();  
            while (toBeRestartedList != null) { 
                startJob(toBeRestartedList);
                toBeRestartedList = toBeRestartedList.nextToBeRestarted;
            }
        } catch (SQLException x) {
            traceException("RESTART-JOBS", x);
        }
    }


    Method getMethod(String className) { 
        Method method = null;
        try { 
            Class cls = Class.forName(className);
            method = cls.getMethod(entryPointMethodName, entryPointMethodSignature);            
        } catch(Exception x) { 
            traceException("LOAD-JOB-TABLE", x);
        }
        return method;
    }

    Method getMethod(Class cls) { 
        Method method = null;
        try { 
            method = cls.getMethod(entryPointMethodName, entryPointMethodSignature);            
        } catch(Exception x) { 
            traceException("LOAD-JOB-TABLE", x);
        }
        return method;
    }

    void loadJobTable() {
        try { 
            Statement stmt = connection.createStatement();
            ResultSet cursor = stmt.executeQuery("select * from JobDesc");
            
            while (cursor.next()) { 
                Job job = new Job(this, jobList);
                job.className = cursor.getString("javaclass");
                job.name = cursor.getString("jobname");
                job.homeDir = cursor.getString("jobhomedir");
                job.hostName = cursor.getString("hostname");
                job.port = cursor.getInt("port");
                job.params = cursor.getString("jobparams");
                job.schedulePolicy = cursor.getInt("schedpolicy"); 
                job.restartCountLimit = cursor.getInt("restartcount"); 
                job.restartDelay = cursor.getInt("restartdelay");
                job.runInterval = cursor.getInt("period"); 
                job.runHourMask = cursor.getInt("hourmask");
                job.runDayOfWeekMask = cursor.getInt("daymask");
                job.method = (job.hostName != null && job.hostName.length() != 0)
                    ? getMethod(RemoteJob.class) : getMethod(job.className);
                job.trace = new FileOutputStream(traceDirectory + job.name + ".out", traceAppend);
                job.log = new PrintStream(job.trace, true);
                jobList = job;
                if (job.method != null) {
                    if (verbose) { 
                        message("Register job '" + job.name + "', method='" + job.method 
                            + "', policy=" + job.schedulePolicy + ", parameters='" + job.params 
                            + "', homedir='" + job.homeDir + "', hostname='" + job.hostName
                            + "' interval=" + job.runInterval + ", restartcount=" + job.restartCountLimit 
                            + ", restartdelay=" + job.restartDelay);
                    }
                    jobs.put(job.name, job);
                }
            }
            cursor.close();

            cursor = stmt.executeQuery("select * from JobFlow order by srcjob");
            Job prevJob = null;
            ArrayList onFinish = new ArrayList();
            ArrayList onFailure = new ArrayList();
            while (cursor.next()) { 
                Job srcJob = getJob(cursor.getString("srcjob"));
                Job dstJob = getJob(cursor.getString("dstjob"));
                if (srcJob != prevJob) { 
                    if (prevJob != null) {
                        prevJob.onFinish = (Job[])onFinish.toArray(new Job[onFinish.size()]);
                        prevJob.onFailure = (Job[])onFailure.toArray(new Job[onFailure.size()]);
                        onFinish.clear();
                        onFailure.clear();
                    }
                    prevJob = srcJob;
                }
                if (cursor.getInt("iffails") != 0) { 
                    onFailure.add(dstJob);
                } else { 
                    onFinish.add(dstJob);
                }
            }
            if (prevJob != null) { 
                prevJob.onFinish = (Job[])onFinish.toArray(new Job[onFinish.size()]);
                prevJob.onFailure = (Job[])onFailure.toArray(new Job[onFailure.size()]);
            }
        } catch (Exception x) { 
            traceException("LOAD-JOB-TABLE", x);
        }
    }

    static final int ONE_HOUR_MSEC = 60*60*1000; 
        
    void startPeriodicJobs() {
        try { 
            Job job;
            PreparedStatement selectStmt = connection.prepareStatement("select * from JobCalendar order by jobname asc, nextrun desc");
            PreparedStatement deleteStmt = connection.prepareStatement("delete from JobCalendar where nextrun < ?");
            long interval = 0;
            long lastCalendarLoadTime = 0;
            Calendar calendar = Calendar.getInstance();

            while (running) {
                System.out.println("interval = " + interval);
                if (interval > 0) { 
                    synchronized (stopEvent) { 
                        stopEvent.wait(interval);
                        if (!running) { 
                            return;
                        }
                    }
                }
                interval = calendarReloadInterval*1000;
                
                long currTime = System.currentTimeMillis();
                calendar.setTimeInMillis(currTime);

                long delta;
                long calendarInterval = interval;

                for (job = jobList; job != null; job = job.next) { 
                    switch (job.schedulePolicy) { 
                      case Job.SCHED_CALENDAR:
                        if (job.nextRun != null && job.nextRun.when.getTime() <= currTime) { 
                            job.nextRun = job.nextRun.next;
                            if (job.nextRun != null && (delta = job.nextRun.when.getTime() - currTime) < calendarInterval) { 
                                calendarInterval = delta;
                            }
                            startJob(job);
                        }
                        break;
                      case Job.SCHED_INTERVAL:
                        if (job.started == null || currTime - job.started.getTime() >= job.runInterval*1000) { 
                            startJob(job);
                        }
                        delta = job.started.getTime() + job.runInterval*1000 - currTime;
                        if (delta < interval) { 
                            interval = delta;
                        }
                        break;
                      case Job.SCHED_DAY_HOUR:
                        if (((1 << calendar.get(Calendar.HOUR_OF_DAY)) & job.runHourMask) != 0
                            && ((1 << calendar.get(Calendar.HOUR_OF_DAY)) & job.proceededHourMask) == 0
                            && ((1 << (calendar.get(Calendar.DAY_OF_WEEK) - 1)) & job.runDayOfWeekMask) != 0)
                        {
                            job.proceededHourMask = 1 << calendar.get(Calendar.HOUR_OF_DAY);
                            startJob(job);
                        }
                    }
                }
                if (currTime - lastCalendarLoadTime >= calendarReloadInterval) {                     
                    lastCalendarLoadTime = currTime;
                    deleteStmt.setTimestamp(1, new Timestamp(currTime));
                    deleteStmt.executeUpdate();
                    for (job = jobList; job != null; job = job.next) { 
                        job.nextRun = null;
                    }
                    ResultSet cursor = selectStmt.executeQuery();
                    job = null;
                    while (cursor.next()) { 
                        String jobName = cursor.getString("jobname");
                        if (job == null || !job.name.equals(jobName)) {
                            job = getJob(jobName);
                        }
                        if (job != null) {
                            Timestamp when = cursor.getTimestamp("nextrun");
                            job.nextRun = new Job.StartTime(when, job.nextRun);
                            delta = when.getTime() - currTime;
                            if (delta < interval) { 
                                interval = delta;
                            }                              
                        }
                    }
                    cursor.close();
                } else if (calendarInterval < interval) {
                    interval = calendarInterval;
                }
                if (interval > ONE_HOUR_MSEC) { 
                    interval = ONE_HOUR_MSEC;
                }
            }
        } catch (Exception x) { 
            traceException("TIMER-THREAD", x);
        }
    }


    Job getJob(String jobName) {
        Job job = (Job)jobs.get(jobName);
        if (job == null) { 
            error("Unknown job '" + jobName + "'");
        }
        return job;
    }

    void error(String msg) { 
        System.err.println(msg);
        if (log != null) { 
            log.println(msg);
        }
    }

    void message(String msg) { 
        System.err.println(msg);
        if (log != null) { 
            log.println(msg);
        }
    }

    void traceException(String jobName, Exception x) {         
        System.err.println("Exception thrown by job '" + jobName + "'");
        x.printStackTrace(System.err);
        if (log != null) { 
            log.println("Exception thrown by job '" + jobName + "'");
            x.printStackTrace(log);
        }
        Job job = (Job)jobs.get(jobName);
        if (job != null) { 
            x.printStackTrace(job.log);
        }
    }

    
    void startQueuePollThread() { 
        if (queuePollInterval != 0) { 
            if (verbose) {
                message("Queue poll thread started with interval " + queuePollInterval);
            }
            new QueuePollThread(this).start();
        }
    }

    public static void main(String[] args) throws Exception { 
        if (args.length != 0) { 
            System.err.println("Usage: java scheduler.Scheduler {property}*");
            System.err.println("Properties:");
            System.err.println("\tverbose:\tExtra trace messages [false]");
            System.err.println("\tdb.url:\tJDBC database URL [jdbc:odbc:LocalServer]");
            System.err.println("\tdb.driver:\tJDBC driver class name [sun.jdbc.odbc.JdbcOdbcDriver]");
            System.err.println("\tdb.user:\tdatabase user name [sa]");
            System.err.println("\tdb.password:\tdatabase user password []");
            System.err.println("\tentry.point:\tname of the method which will be invoked by scheduler [run]");
            System.err.println("\ttrace.dir:\tdirectory where trace files for local tasks will be created [.]");
            System.err.println("\ttrace.append:\twhether to append or rewrite trace file for local tasks [false]");
            System.err.println("\tlog.file.path:\tpath to the scheduler log file [scheduler.log]");
            System.err.println("\tlog.append:\twhether to append or rewrite scheduler log file [false]");
            System.err.println("\tqueue.poll.interval:\tinterval in seconds of polling state of queues [60]");
            System.err.println("\tqueue.create:\twhether scheduler should create queues [false]");
            System.err.println("\tipc.server.name:\tname of the host at which IPC server is running [localhost]");
            System.err.println("\tipc.server.port:\tport listened by IPC server [6000]");
            return;
        }
        instance = new Scheduler();
        Thread schedulerThread = new Thread(instance);
        schedulerThread.start();

        System.out.println("Java Job Scheduler: type exit to terminate...");
        byte[] inputBuffer = new byte[256];
        String cmd;
        do { 
            System.out.print("> ");
            int len = System.in.read(inputBuffer);
            cmd = new String(inputBuffer, 0, len).trim();
        } while (!cmd.equalsIgnoreCase("exit"));

        instance.terminate();
    }
    
    public static Scheduler getScheduler() {
        return instance;
    }

    public Scheduler() throws Exception { 
        String dbUrl = System.getProperty("db.url", "jdbc:odbc:LocalServer");
        String dbUser = System.getProperty("db.user", "sa");
        String dbPassword = System.getProperty("db.password", "");
        String dbDriver = System.getProperty("db.driver", "sun.jdbc.odbc.JdbcOdbcDriver");
        Class.forName(dbDriver);

        String logFilePath = System.getProperty("log.file.path", "scheduler.log");
        boolean logAppend = Boolean.getBoolean("log.append");
        if (logFilePath.length() != 0) {
            log = new PrintStream(new FileOutputStream(logFilePath, logAppend), true);
        }

        entryPointMethodName = System.getProperty("entry.point", "run");

        queuePollInterval = Integer.parseInt(System.getProperty("queue.poll.interval", "60"), 10);
        calendarReloadInterval = Integer.parseInt(System.getProperty("queue.poll.interval", "60"), 10);
        ipcServerName = System.getProperty("ipc.server.name", "localhost");
        ipcServerPort = Integer.parseInt(System.getProperty("ipc.server.port", "6000"), 10);
        traceDirectory = System.getProperty("trace.dir", ".");
        if (!traceDirectory.endsWith(File.separator)) { 
            traceDirectory += File.separatorChar;
        }
        verbose = Boolean.getBoolean("verbose");
        createQueues = Boolean.getBoolean("queue.create");
        traceAppend = Boolean.getBoolean("trace.append");        
        connection = DriverManager.getConnection(dbUrl, dbUser, dbPassword);	
        connection.setAutoCommit(true);
        updateJobStmt = connection.prepareStatement("update JobState set status=?, started=?, finished=? where jobname=?");

        pool = ThreadPool.getInstance();
        jobs = new HashMap();
        
        stopEvent = new Object();
        running = true;
    }

    Connection  connection;
    ThreadPool  pool;
    HashMap     jobs;
    Job         jobList;

    PrintStream log;
    String      ipcServerName;
    int         ipcServerPort;
    int         queuePollInterval;
    int         calendarReloadInterval;

    String      traceDirectory;
    boolean     traceAppend;
    boolean     running;
    boolean     verbose;
    boolean     createQueues;
    Object      stopEvent;
    String      entryPointMethodName;
    static final Class[] entryPointMethodSignature = { Job.class };

    PreparedStatement updateJobStmt;
    static Scheduler instance;

};










